//
// This file is automatically generated by generate_auto_job.
// Edits will be automatically overwritten if the build is
// run in maintainer mode.
//
// clang-format off
//
QPDF_DLL CopyAttConfig* prefix(std::string const& parameter);
QPDF_DLL CopyAttConfig* password(std::string const& parameter);
