//
// This file is automatically generated by generate_auto_job.
// Edits will be automatically overwritten if the build is
// run in maintainer mode.
//
// clang-format off
//
QPDF_DLL Config* allowWeakCrypto();
QPDF_DLL Config* check();
QPDF_DLL Config* checkLinearization();
QPDF_DLL Config* coalesceContents();
QPDF_DLL Config* decrypt();
QPDF_DLL Config* deterministicId();
QPDF_DLL Config* externalizeInlineImages();
QPDF_DLL Config* filteredStreamData();
QPDF_DLL Config* flattenRotation();
QPDF_DLL Config* generateAppearances();
QPDF_DLL Config* ignoreXrefStreams();
QPDF_DLL Config* isEncrypted();
QPDF_DLL Config* jsonInput();
QPDF_DLL Config* keepInlineImages();
QPDF_DLL Config* linearize();
QPDF_DLL Config* listAttachments();
QPDF_DLL Config* newlineBeforeEndstream();
QPDF_DLL Config* noOriginalObjectIds();
QPDF_DLL Config* noWarn();
QPDF_DLL Config* optimizeImages();
QPDF_DLL Config* passwordIsHexKey();
QPDF_DLL Config* preserveUnreferenced();
QPDF_DLL Config* preserveUnreferencedResources();
QPDF_DLL Config* progress();
QPDF_DLL Config* qdf();
QPDF_DLL Config* rawStreamData();
QPDF_DLL Config* recompressFlate();
QPDF_DLL Config* removeAcroform();
QPDF_DLL Config* removeInfo();
QPDF_DLL Config* removeMetadata();
QPDF_DLL Config* removePageLabels();
QPDF_DLL Config* removeStructure();
QPDF_DLL Config* reportMemoryUsage();
QPDF_DLL Config* requiresPassword();
QPDF_DLL Config* removeRestrictions();
QPDF_DLL Config* showEncryption();
QPDF_DLL Config* showEncryptionKey();
QPDF_DLL Config* showLinearization();
QPDF_DLL Config* showNpages();
QPDF_DLL Config* showPages();
QPDF_DLL Config* showXref();
QPDF_DLL Config* staticAesIv();
QPDF_DLL Config* staticId();
QPDF_DLL Config* suppressPasswordRecovery();
QPDF_DLL Config* suppressRecovery();
QPDF_DLL Config* testJsonSchema();
QPDF_DLL Config* verbose();
QPDF_DLL Config* warningExit0();
QPDF_DLL Config* withImages();
QPDF_DLL Config* compressionLevel(std::string const& parameter);
QPDF_DLL Config* jpegQuality(std::string const& parameter);
QPDF_DLL Config* copyEncryption(std::string const& parameter);
QPDF_DLL Config* encryptionFilePassword(std::string const& parameter);
QPDF_DLL Config* forceVersion(std::string const& parameter);
QPDF_DLL Config* iiMinBytes(std::string const& parameter);
QPDF_DLL Config* jobJsonFile(std::string const& parameter);
QPDF_DLL Config* jsonObject(std::string const& parameter);
QPDF_DLL Config* keepFilesOpenThreshold(std::string const& parameter);
QPDF_DLL Config* linearizePass1(std::string const& parameter);
QPDF_DLL Config* minVersion(std::string const& parameter);
QPDF_DLL Config* oiMinArea(std::string const& parameter);
QPDF_DLL Config* oiMinHeight(std::string const& parameter);
QPDF_DLL Config* oiMinWidth(std::string const& parameter);
QPDF_DLL Config* password(std::string const& parameter);
QPDF_DLL Config* passwordFile(std::string const& parameter);
QPDF_DLL Config* removeAttachment(std::string const& parameter);
QPDF_DLL Config* rotate(std::string const& parameter);
QPDF_DLL Config* showAttachment(std::string const& parameter);
QPDF_DLL Config* showObject(std::string const& parameter);
QPDF_DLL Config* jsonStreamPrefix(std::string const& parameter);
QPDF_DLL Config* updateFromJson(std::string const& parameter);
QPDF_DLL Config* collate(std::string const& parameter);
QPDF_DLL Config* collate();
QPDF_DLL Config* splitPages(std::string const& parameter);
QPDF_DLL Config* splitPages();
QPDF_DLL Config* compressStreams(std::string const& parameter);
QPDF_DLL Config* decodeLevel(std::string const& parameter);
QPDF_DLL Config* flattenAnnotations(std::string const& parameter);
QPDF_DLL Config* jsonKey(std::string const& parameter);
QPDF_DLL Config* jsonStreamData(std::string const& parameter);
QPDF_DLL Config* keepFilesOpen(std::string const& parameter);
QPDF_DLL Config* normalizeContent(std::string const& parameter);
QPDF_DLL Config* objectStreams(std::string const& parameter);
QPDF_DLL Config* passwordMode(std::string const& parameter);
QPDF_DLL Config* removeUnreferencedResources(std::string const& parameter);
QPDF_DLL Config* streamData(std::string const& parameter);
QPDF_DLL Config* json(std::string const& parameter);
QPDF_DLL Config* json();
QPDF_DLL Config* jsonOutput(std::string const& parameter);
QPDF_DLL Config* jsonOutput();
