//
// This file is automatically generated by generate_auto_job.
// Edits will be automatically overwritten if the build is
// run in maintainer mode.
//
// clang-format off
//
QPDF_DLL EncConfig* extract(std::string const& parameter);
QPDF_DLL EncConfig* annotate(std::string const& parameter);
QPDF_DLL EncConfig* print(std::string const& parameter);
QPDF_DLL EncConfig* modify(std::string const& parameter);
QPDF_DLL EncConfig* cleartextMetadata();
QPDF_DLL EncConfig* forceV4();
QPDF_DLL EncConfig* accessibility(std::string const& parameter);
QPDF_DLL EncConfig* assemble(std::string const& parameter);
QPDF_DLL EncConfig* form(std::string const& parameter);
QPDF_DLL EncConfig* modifyOther(std::string const& parameter);
QPDF_DLL EncConfig* useAes(std::string const& parameter);
QPDF_DLL EncConfig* forceR5();
QPDF_DLL EncConfig* allowInsecure();
